/*
 * FILE NAME Gemini_gpio.c
 *
 * BRIEF MODULE DESCRIPTION
 *  API for gemini GPIO module
 *  Driver for gemini GPIO module
 *
 *  Author: StorLink, Corp.
 *          Jason Lee <jason@storlink.com.tw>
 *
 * Copyright 2005 StorLink, Corp.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMit8712D  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE	LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMit8712D   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, writ8712  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/arch/sl2312.h>
#include <asm/arch/irqs.h>
#include <asm/arch/gemini_gpio.h>

#define GEMINI_GPIO_BASE1		SL2312_GPIO_BASE
#define GEMINI_GPIO_BASE2		SL2312_GPIO_BASE1
#define GEMINI_GPIO_BASE3		SL2312_GPIO_BASE2

#define BANK1_BIT(__BIT__)		(0+__BIT__)
#define BANK2_BIT(__BIT__)		(32+__BIT__)
#define BANK3_BIT(__BIT__)		(64+__BIT__)

#define GPIO_SET	3
#define MAX_GPIO_LINE	32*GPIO_SET

wait_queue_head_t gemini_gpio_wait[MAX_GPIO_LINE];

enum GPIO_REG
{
    GPIO_DATA_OUT   		= 0x00,
    GPIO_DATA_IN    		= 0x04,
    GPIO_PIN_DIR    		= 0x08,
    GPIO_BY_PASS    		= 0x0C,
    GPIO_DATA_SET   		= 0x10,
    GPIO_DATA_CLEAR 		= 0x14,
    GPIO_PULL_ENABLE 		= 0x18,
    GPIO_PULL_TYPE 			= 0x1C,
    GPIO_INT_ENABLE 		= 0x20,
    GPIO_INT_RAW_STATUS 	= 0x24,
    GPIO_INT_MASK_STATUS 	= 0x28,
    GPIO_INT_MASK 			= 0x2C,
    GPIO_INT_CLEAR 			= 0x30,
    GPIO_INT_TRIG 			= 0x34,
    GPIO_INT_BOTH 			= 0x38,
    GPIO_INT_POLAR 			= 0x3C
};

unsigned int regist_gpio_int0=0,regist_gpio_int1=0,regist_gpio_int2=0;
static int gemini_gpio_debug=0;
#define DEB(x)  if (gemini_gpio_debug>=1) x

struct gpio_irqaction {
	void (*handler)(unsigned int);
};

struct gpio_irqaction irq_handle[MAX_GPIO_LINE];


void gemini_gpio_access(unsigned char pin, unsigned char bit, unsigned char addr_offset)
{
	unsigned char set = pin >>5;		// each GPIO set has 32 pins
	unsigned int status,addr;

	if (set==0)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE1);
	else if (set==1)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE2);
	else
		addr = IO_ADDRESS(GEMINI_GPIO_BASE3);
	addr+=addr_offset;
	status = readl(addr);

	status &= ~(1 << (pin %32));
	status |= (bit << (pin % 32));
	writel(status,addr);
}

// set: 1-2
// mode:0: input, 1: output
void set_gemini_gpio_io_mode(unsigned char pin, unsigned char mode)
{
	unsigned char set = pin >>5;		// each GPIO set has 32 pins
	unsigned int status,addr;

	if (set==0)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE1)+GPIO_PIN_DIR;
	else if (set==1)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE2)+GPIO_PIN_DIR;
	else
		addr = IO_ADDRESS(GEMINI_GPIO_BASE3)+GPIO_PIN_DIR;
	status = readl(addr);

	status &= ~(1 << (pin %32));
	status |= (mode << (pin % 32));
	writel(status,addr);
}

// set: 1-2
// high: 1:high, 0:low
void set_gemini_gpio_pin_status(unsigned char pin, unsigned char high)
{
	unsigned char set = pin >>5;		// each GPIO set has 32 pins
	unsigned int status,addr;
	
	if (set==0)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE1);
	else if (set==1)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE2);
	else
		addr = IO_ADDRESS(GEMINI_GPIO_BASE3);
	addr+=GPIO_DATA_OUT;

	status = readl(addr);
	if (high==0)
		status&=~(1 << (pin % 32));
	else
		status|=(1 << (pin % 32));
	writel(status,addr);

	DEB(printk("set_gemini_gpio_pin_status: write 0x%X at 0x%X.\n",status,addr));
}

void set_gemini_gpio_fan_status(unsigned char pin, unsigned char value)
{
	unsigned int status,addr;

	addr=IO_ADDRESS(GEMINI_GPIO_BASE3);
	addr+=GPIO_DATA_OUT;

	status = readl(addr);
	status&=~(0x1F << 8);
	status|=(value << 8);
	writel(status,addr);

// don't want to see that, by Freecom
//	DEB(printk("set_gemini_gpio_fan_status: write 0x%X at 0x%X.\n",status,addr));
}

// set: 1-2
// return: 1:high, 0:low
int get_gemini_gpio_pin_status(unsigned char pin)
{
	unsigned char set = pin >>5;		// each GPIO set has 32 pins
	unsigned int status,addr;
	
	if (set==0)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE1)+GPIO_DATA_IN;
	else if (set==1)
		addr = IO_ADDRESS(GEMINI_GPIO_BASE2)+GPIO_DATA_IN;
	else
		addr = IO_ADDRESS(GEMINI_GPIO_BASE3)+GPIO_DATA_IN;

	status = readl(addr);

	DEB(printk("get_gemini_gpio_pin_status pin %u - 0x%X - address: 0x%X.\n",pin,status,addr));

	return (status&(1<<(pin % 32)))?1:0; 
}


// wait interrupt event
int wait_gemini_gpio_pin_interrupt(struct gemini_gpio_ioctl_data *ioctl_data)
{
		
	DEB(printk("wait GPIO_%d Interrupt\n",ioctl_data->pin));
	if(ioctl_data->use_default==1)
		init_gpio_int(ioctl_data->pin, 0, 0, 0);
	else
		init_gpio_int(ioctl_data->pin, ioctl_data->trig_type, ioctl_data->trig_polar, \
					ioctl_data->trig_both);
		
	interruptible_sleep_on(&gemini_gpio_wait[ioctl_data->pin]);
	return 0;
}

void init_gpio_int(__u32 pin,__u8 trig_type,__u8 trig_polar,__u8 trig_both)
{
	unsigned int data =0,addr,base;
	unsigned char set = pin >>5;	
	
	if (set==0)
		base = IO_ADDRESS(GEMINI_GPIO_BASE1);
	else if (set==1)
		base = IO_ADDRESS(GEMINI_GPIO_BASE2);
	else
		base = IO_ADDRESS(GEMINI_GPIO_BASE3);
	
	addr = base + GPIO_INT_TRIG ;
	data = readl(addr);
//	data &= ~(1<<(pin%32)) ;			// edge trig
	trig_type ? (data|=(1<<(pin%32))):(data &= ~(1<<(pin%32)));
	writel(data,addr);
	
	addr = base + GPIO_INT_POLAR ;
	data = readl(addr);
//	data &= ~(1<<(pin%32));				// rising edge
	trig_polar ? (data|=(1<<(pin%32))):(data &= ~(1<<(pin%32)));
	writel(data,addr);
	
	addr = base + GPIO_INT_BOTH ;
	data = readl(addr);
//	data &= ~(1<<(pin%32)) ;			// single edge
	trig_both ? (data|=(1<<(pin%32))):(data &= ~(1<<(pin%32)));
	writel(data,addr);
	
	addr = base + GPIO_INT_MASK ;
	data = readl(addr);
	data &= ~(1<<(pin%32));				// unmask
	writel(data,addr);
	
	addr = base + GPIO_INT_CLEAR ;
	data = readl(addr);
	data |= 1<<(pin%32) ;				// Clear interrupt before Enable it
	writel(data,addr);	
	
	addr = base + GPIO_INT_ENABLE ;
	data = readl(addr);
	data |= 1<<(pin%32) ;				// Enable interrupt
	writel(data,addr);	

}

static irqreturn_t gpio_int1(void)
{
	unsigned int i,int_src,addr,data,base,int_status;
	unsigned int mask_stat;
	
	base = IO_ADDRESS(GEMINI_GPIO_BASE1);
	addr = base + GPIO_INT_RAW_STATUS ;
	int_status = int_src = readl(addr);
	
	addr = base + GPIO_INT_MASK;
	mask_stat = readl(addr);
	
	int_src &= ~mask_stat;							// ignore INT that was masked
	for(i=0;int_src;int_src>>=1,i++) {
		if((int_src&(1))==0)
			continue;
			
		addr = base + GPIO_INT_MASK ;
		data = readl(addr);
		data |= (1<<i);								// mask INT
		writel(data,addr);
		if( (regist_gpio_int0&(1<<i))&&irq_handle[i].handler ){
			irq_handle[i].handler(int_status);
		}
		else{
			wake_up(&gemini_gpio_wait[i]);				// wake up 
			DEB(printk("wake up gpio %d",i));
		}
		addr = base + GPIO_INT_CLEAR ;
		data = readl(addr);
		data |= (1<<i) ;								// clear INT
		writel(data,addr);
		
		addr = base + GPIO_INT_MASK ;
		data = readl(addr);
		data &= ~(1<<i);								// unmask INT
		writel(data,addr);
		
	}

	return IRQ_HANDLED;
}

static irqreturn_t gpio_int2(void)
{
	unsigned int i,int_src,addr,data,base,int_status;
	unsigned int mask_stat;
	
	base = IO_ADDRESS(GEMINI_GPIO_BASE2);
	addr = base + GPIO_INT_RAW_STATUS ;
	int_status = int_src = readl(addr);
	
	addr = base + GPIO_INT_MASK;
	mask_stat = readl(addr);
	
	int_src &= ~mask_stat;							// ignore INT that was masked
	for(i=0;int_src;int_src>>=1,i++) {
		if((int_src&(1))==0)
			continue;
			
		addr = base + GPIO_INT_MASK ;
		data = readl(addr);
		data |= (1<<i);								// mask INT
		writel(data,addr);
		if( (regist_gpio_int1&(1<<i))&&irq_handle[i+32].handler ){
			irq_handle[i+32].handler(int_status);
		}
		else{
			wake_up(&gemini_gpio_wait[i+32]);				// wake up 
			DEB(printk("wake up gpio %d",i+32));
		}
		addr = base + GPIO_INT_CLEAR ;
		data = readl(addr);
		data |= (1<<i) ;								// clear INT
		writel(data,addr);
		
		addr = base + GPIO_INT_MASK ;
		data = readl(addr);
		data &= ~(1<<i);								// unmask INT
		writel(data,addr);
		
	}

	return IRQ_HANDLED;
}

static irqreturn_t gpio_int3(void)
{
	unsigned int i,int_src,addr,data,base,int_status;
	unsigned int mask_stat;
	
	base = IO_ADDRESS(GEMINI_GPIO_BASE3);
	addr = base + GPIO_INT_RAW_STATUS ;
	int_status = int_src = readl(addr);
	
	addr = base + GPIO_INT_MASK;
	mask_stat = readl(addr);
	
	int_src &= ~mask_stat;							// ignore INT that was masked
	for(i=0;int_src;int_src>>=1,i++) {
		if((int_src&(1))==0)
			continue;
			
		addr = base + GPIO_INT_MASK ;
		data = readl(addr);
		data |= (1<<i);								// mask INT
		writel(data,addr);
		if( (regist_gpio_int2&(1<<i))&&irq_handle[i+64].handler ){
			irq_handle[i+64].handler(int_status);
		}
		else{
			wake_up(&gemini_gpio_wait[i+64]);				// wake up 
			DEB(printk("wake up gpio %d",i+64));
		}
		addr = base + GPIO_INT_CLEAR ;
		data = readl(addr);
		data |= (1<<i) ;								// clear INT
		writel(data,addr);
		
		addr = base + GPIO_INT_MASK ;
		data = readl(addr);
		data &= ~(1<<i);								// unmask INT
		writel(data,addr);
		
	}

	return IRQ_HANDLED;
}

static int gemini_gpio_open(struct inode *inode, struct file *file)
{
	return 0;
}


static int gemini_gpio_release(struct inode *inode, struct file *file)
{
	return 0;
}


static int gemini_gpio_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	struct gemini_gpio_ioctl_data ioctl_data;
	
	if (copy_from_user(&ioctl_data, (struct gemini_gpio_ioctl_data *)arg, sizeof(ioctl_data)))
		return -EFAULT;
	
	if (ioctl_data.pin >= MAX_GPIO_LINE) 	
	{
		printk("gemini_gpio ioctl fail[pin > %d]\n",MAX_GPIO_LINE);
		return -EFAULT;
	}
// don't want to see that, by Freecom
//	DEB(printk("gemini_gpio ioctl successfully\n"));

	switch(cmd) {
		case GEMINI_SET_GPIO_PIN_DIR:			// Set pin Direction
			set_gemini_gpio_io_mode(ioctl_data.pin, ioctl_data.status);
			break;

		case GEMINI_SET_GPIO_PIN_STATUS:		// Set pin Status
			set_gemini_gpio_pin_status(ioctl_data.pin, ioctl_data.status);
			break;

		case GEMINI_GET_GPIO_PIN_STATUS:		// Get pin Status
			ioctl_data.status = get_gemini_gpio_pin_status(ioctl_data.pin);
			if (copy_to_user((struct gemini_gpio_ioctl_data *)arg,&ioctl_data, sizeof(ioctl_data))) 
				return -EFAULT;
			break;

		case GEMINI_WAIT_GPIO_PIN_INT:			// Wait pin Interrupt
			if (ioctl_data.pin<32)
			{
				if ((regist_gpio_int0 & (1<<ioctl_data.pin))!=0)
					return -EBUSY;
			}
			else if (ioctl_data.pin<64)
			{
				if ((regist_gpio_int1 & (1<<(ioctl_data.pin-32)))!=0)
					return -EBUSY;
			}
			else
			{
				if ((regist_gpio_int2 & (1<<(ioctl_data.pin-64)))!=0)
					return -EBUSY;
			}

			return wait_gemini_gpio_pin_interrupt(&ioctl_data);
			
		case GEMINI_SET_GPIO_FAN_STATUS:		// Set fan Status
			set_gemini_gpio_fan_status(ioctl_data.pin, ioctl_data.status);
			break;

		default:
			return -ENOIOCTLCMD;

	}
	return 0;
}

/************************************************************************/
/* request_gpio_irq()													*/
/* args: 	bit 	--> the pin number of gpio to regist interrupt		*/
/*			handler	--> handle of interrupt service routine				*/
/*			level	--> trigger type level/edge							*/
/*			high	--> trigger mode high/low or rigsing/falling		*/
/*			both	--> if edge trigger single/both edge				*/
/* return: 0 if succeed													*/
/*	   	   -EINVAL indicate this gpio was assigned to other module		*/
/************************************************************************/
int request_gpio_irq(int bit,void (*handler)(int),char level,char high,char both)
{
	if( (bit>=MAX_GPIO_LINE)||(bit<0) )
		return -EINVAL;
	
	if(bit<32){
		if(regist_gpio_int0&(1<<bit))
			return -EUSERS;
		regist_gpio_int0 |= 1<<bit;
	}
	else if(bit<64){
		if(regist_gpio_int1&(1<<(bit-32)))
			return -EUSERS;
		regist_gpio_int1 |= 1<<(bit-32);
	}
	else{
		if(regist_gpio_int2&(1<<(bit-64)))
			return -EUSERS;
		regist_gpio_int2 |= 1<<(bit-64);
	}
	
	irq_handle[bit].handler = (void *)handler;
	init_gpio_int(bit,level,high,both);
	
	return 0;
}
//EXPORT_SYMBOL(request_gpio_irq);

/************************************************************************/
/* free_gpio_irq()														*/
/* args:	bit --> pin number of gpio to release irq					*/
/* return: None															*/
/************************************************************************/
int free_gpio_irq(int bit)
{
	unsigned int base,addr,data;
	
	if( (bit>=MAX_GPIO_LINE)||(bit<0) )
		return -EINVAL;

	if (bit<32)
		base=IO_ADDRESS(GEMINI_GPIO_BASE1);
	else if (bit<64)
		base=IO_ADDRESS(GEMINI_GPIO_BASE2);
	else
		base=IO_ADDRESS(GEMINI_GPIO_BASE3);

	irq_handle[bit].handler = NULL;
	
	if(bit<32)
		regist_gpio_int0 &= ~(1<<bit);
	else if(bit<64){
		bit -= 32;
		regist_gpio_int1 &= ~(1<<bit);
	}
	else{
		bit -= 64;
		regist_gpio_int2 &= ~(1<<bit);
	}
	
	addr = base + GPIO_INT_MASK ;
	data = readl(addr);
	data |= 1<<(bit);								// mask INT
	writel(data,addr);
	return 0;
}
//EXPORT_SYMBOL(request_gpio_irq);

static struct file_operations gemini_gpio_fops = {
	.owner	=	THIS_MODULE,
	.ioctl	=	gemini_gpio_ioctl,
	.open	=	gemini_gpio_open,
	.release=	gemini_gpio_release,
};

/* GPIO_MINOR in include/linux/miscdevice.h */
static struct miscdevice gemini_gpio_miscdev =
{
	GPIO_MINOR,
	"sl_gpio",
	&gemini_gpio_fops
};

static  struct work_struct work_item;

static long unsigned int resetJiffies;

static void work_item_function(void *p)
{
	char *argv[3] = { NULL, NULL, NULL };
	char *envp[5] = { NULL, NULL, NULL, NULL, NULL };
	char *action_str[32];
	int ret;

	DEB(printk("work_item_function: %s.\n",(char *)p));

	sprintf((char *)action_str,"ACTION=%s",(char *)p);

	argv[0] = "/sbin/hotplug";
	argv[1] = "key";

	envp [0] = "HOME=/";
	envp [1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
	envp [2] = "INTERFACE=";
	envp [3] = (char *)action_str;

	ret=call_usermodehelper(argv[0],argv,envp,1);
	DEB(printk ("work_item_function: %s result: %d.\n",(char *)action_str,ret));
}

static void irq_button_reset_handler(int status)
{
	mdelay(50);

	DEB(printk("irq_button_reset_handler. status: 0x%X.\n",get_gemini_gpio_pin_status(BANK3_BIT(20))));
	DEB(printk("irq_button_reset_handler. jiffies: 0x%08lX.\n",jiffies));
	DEB(printk("irq_button_reset_handler. resetJiffies: 0x%08lX.\n",resetJiffies));

	if (get_gemini_gpio_pin_status(BANK3_BIT(20))==0)
	{
		resetJiffies=jiffies;
	}
	else
	{
		if (abs(jiffies-resetJiffies)>=4*HZ)
		{
			INIT_WORK(&work_item,work_item_function,"DEFAULTS");
		}
		else
		{
			INIT_WORK(&work_item,work_item_function,"REBOOT");
		}

		schedule_work(&work_item);
	}
}

static void irq_button_usb1_handler(int status)
{
	mdelay(50);

	DEB(printk("irq_button_usb1_handler. status: 0x%X.\n",get_gemini_gpio_pin_status(BANK3_BIT(16))));

	if (get_gemini_gpio_pin_status(BANK3_BIT(16))==0)
	{
		INIT_WORK(&work_item,work_item_function,"UNPLUG USB1");
		schedule_work(&work_item);
	}
}

static void irq_button_usb2_handler(int status)
{
	mdelay(50);

	DEB(printk("irq_button_usb2_handler. status: 0x%X.\n",get_gemini_gpio_pin_status(BANK3_BIT(17))));

	if (get_gemini_gpio_pin_status(BANK3_BIT(17))==0)
	{
		INIT_WORK(&work_item,work_item_function,"UNPLUG USB2");
		schedule_work(&work_item);
	}
}

static void irq_button_user_handler(int status)
{
	mdelay(50);

	DEB(printk("irq_button_user_handler. status: 0x%X.\n",get_gemini_gpio_pin_status(BANK3_BIT(18))));

	if (get_gemini_gpio_pin_status(BANK3_BIT(18))==0)
	{
		INIT_WORK(&work_item,work_item_function,"USEREVENT");
		schedule_work(&work_item);
	}
}

static void irq_button_power_handler(int status)
{
	mdelay(50);

	DEB(printk("irq_button_power_handler. status: 0x%X.\n",get_gemini_gpio_pin_status(BANK3_BIT(19))));

	if (get_gemini_gpio_pin_status(BANK3_BIT(19))==0)
	{
		INIT_WORK(&work_item,work_item_function,"SHUTDOWN");
		schedule_work(&work_item);
	}
}

int __init gemini_gpio_init(void)
{
	int i,ret=0;

	misc_register(&gemini_gpio_miscdev);
	printk("Freecom/Gemini GPIO Init.\n");

	for(i=0;i<MAX_GPIO_LINE;i++)
		irq_handle[i].handler = NULL;	// init handle pointer
		
	ret = request_irq(IRQ_GPIO, (void *)gpio_int1, SA_INTERRUPT, "GPIO1", NULL);
	if (ret)
		printk("Error: Register IRQ for GPIO1:%d\n",ret);
	ret = request_irq(IRQ_GPIO1, (void *)gpio_int2, SA_INTERRUPT, "GPIO2", NULL);
	if (ret)
		printk("Error: Register IRQ for GPIO2:%d\n",ret);
	ret = request_irq(IRQ_GPIO2, (void *)gpio_int3, SA_INTERRUPT, "GPIO3", NULL);
	if (ret)
		printk("Error: Register IRQ for GPIO3:%d\n",ret);

	for (i = 0; i < MAX_GPIO_LINE; i++) {
		init_waitqueue_head(&gemini_gpio_wait[i]);
	}

	//
	// button stuff
	//

	// usb1 button
	gemini_gpio_access(BANK3_BIT(16),0,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(16),1,GPIO_PULL_ENABLE);
	gemini_gpio_access(BANK3_BIT(16),0,GPIO_PULL_TYPE);
	request_gpio_irq(BANK3_BIT(16),irq_button_usb1_handler,0,1,1);

	// usb2 button
	gemini_gpio_access(BANK3_BIT(17),0,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(17),1,GPIO_PULL_ENABLE);
	gemini_gpio_access(BANK3_BIT(17),0,GPIO_PULL_TYPE);
	request_gpio_irq(BANK3_BIT(17),irq_button_usb2_handler,0,1,1);

	// sync button
	gemini_gpio_access(BANK3_BIT(18),0,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(18),1,GPIO_PULL_ENABLE);
	gemini_gpio_access(BANK3_BIT(18),0,GPIO_PULL_TYPE);
	request_gpio_irq(BANK3_BIT(18),irq_button_user_handler,0,1,1);

	// power button
	gemini_gpio_access(BANK3_BIT(19),0,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(19),1,GPIO_PULL_ENABLE);
	gemini_gpio_access(BANK3_BIT(19),0,GPIO_PULL_TYPE);
	request_gpio_irq(BANK3_BIT(19),irq_button_power_handler,0,1,1);

	// reset button
	resetJiffies=jiffies;
	gemini_gpio_access(BANK3_BIT(20),0,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(20),1,GPIO_PULL_ENABLE);
	gemini_gpio_access(BANK3_BIT(20),0,GPIO_PULL_TYPE);
	request_gpio_irq(BANK3_BIT(20),irq_button_reset_handler,0,1,1);

	//
	// led stuff
	//

	// turn usb1 led off
	gemini_gpio_access(BANK3_BIT(24),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(24),0,GPIO_DATA_OUT);
	// turn usb2 led off
	gemini_gpio_access(BANK3_BIT(25),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(25),0,GPIO_DATA_OUT);
	// turn sync led on
	gemini_gpio_access(BANK3_BIT(26),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(26),1,GPIO_DATA_OUT);
	// turn power led on
	gemini_gpio_access(BANK3_BIT(27),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(27),1,GPIO_DATA_OUT);
	// turn access led off
	gemini_gpio_access(BANK3_BIT(28),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(28),0,GPIO_DATA_OUT);

	//
	// fan stuff
	//

	gemini_gpio_access(BANK3_BIT(8),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(8),0,GPIO_DATA_OUT);
	gemini_gpio_access(BANK3_BIT(9),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(9),0,GPIO_DATA_OUT);
	gemini_gpio_access(BANK3_BIT(10),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(10),0,GPIO_DATA_OUT);
	gemini_gpio_access(BANK3_BIT(11),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(11),1,GPIO_DATA_OUT);	// 75%
	gemini_gpio_access(BANK3_BIT(12),1,GPIO_PIN_DIR);
	gemini_gpio_access(BANK3_BIT(12),1,GPIO_DATA_OUT);	// 75%

	return 0;
}	

void __exit gemini_gpio_exit(void)
{
	free_irq(IRQ_GPIO1,NULL);
	free_irq(IRQ_GPIO2,NULL);
	misc_deregister(&gemini_gpio_miscdev);
}

module_init(gemini_gpio_init);
module_exit(gemini_gpio_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jason Lee <jason@storlink.com.tw>");
MODULE_DESCRIPTION("Storlink GPIO driver");
